/* screenblank.c - screen blanking daemon
**
** Comments to:
**   jef@acme.com
**   jef@ee.lbl.gov
**   jef@well.com
**
** Improved activity check courtesy of Paul Eggert.
**
** Copyright (C) 1989, 1990, 1991, 1993, 1994, 1995 by Jef Poskanzer.
**
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted, provided
** that the above copyright notice appear in all copies and that both that
** copyright notice and this permission notice appear in supporting
** documentation.  This software is provided "as is" without express or
** implied warranty.
*/

#undef DEBUG

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <syslog.h>
#ifndef NO_POLL
#include <poll.h>
#endif /* NO_POLL */
#ifdef SUN_FBIO
#include <sun/fbio.h>
#else /* SUN_FBIO */
#include <sys/fbio.h>
#endif /* SUN_FBIO */
#ifndef NO_DEVFBS
#include <dirent.h>
#endif /* NO_DEVFBS */
#ifndef NO_POWERCTL
#include <sys/cg14io.h>
#ifndef SET_MONITOR_POWER
#define NO_POWERCTL
#endif /* SET_MONITOR_POWER */
#endif /* NO_POWERCTL */


/* Definitions. */

#define CONSOLE_NAME "/dev/console"
#define KEYBOARD_NAME "/dev/kbd"
#define MOUSE_NAME "/dev/mouse"

#define DEFAULT_BLANK_SECS 600		/* ten minutes */

#ifndef NO_POWERCTL
#define DEFAULT_POWER_SECS 3600		/* one hour */
#endif /* NO_POWERCTL */

#define IDLECHECK_MSECS 250		/* four times a second */

#define MAX_FBS 100

#define min(a,b) ( (int) (a) < (int) (b) ? (a) : (b) )
#define max(a,b) ( (int) (a) > (int) (b) ? (a) : (b) )


/* Forward routines. */

static void find_fbs();
static void check_fbs();
static void daemonize();
static char* resolve_symlink();
static void main_loop();
static unsigned idle_time();
static void msleep();
static int my_open();
static void my_stat();
static void blank();
static void unblank();
static void set_video();
#ifndef NO_POWERCTL
static void poweroff();
static void poweron();
static void set_power();
#endif /* NO_POWERCTL */
static void sig_handle();
static void die();
#ifdef NO_STRERROR
static char* strerror();
#endif /* NO_STRERROR */


/* Variables. */

static char* argv0;
static int ign_mouse;
static int ign_keyboard;
static char* console_name;
static char* keyboard_name;
static char* mouse_name;
static int num_fbs;
static char* fbnames[MAX_FBS + 1];
static int fb_fd[MAX_FBS];
static dev_t fb_dev[MAX_FBS];
static ino_t fb_ino[MAX_FBS];
static unsigned blank_secs;
static int video_state[MAX_FBS];
static int can_setvideo[MAX_FBS];
#ifndef NO_POWERCTL
static unsigned power_secs;
static int power_state[MAX_FBS];
static int can_setpower[MAX_FBS];
#endif /* NO_POWERCTL */
static int console_fd;
static int keyboard_fd;
static int mouse_fd;
static int some_opens_failed;


/* Routines. */

void
main( argc, argv )
    int argc;
    char* argv[];
    {
    int argn, num_fbnames;
    char* usage = "usage:  %s [-mouse] [-keyboard] [-delay <seconds>] [-fb <framebuffer> ...]\n";

    argv0 = *argv;

    /* Parse flags. */
    argn = 1;
    ign_mouse = 0;
    ign_keyboard = 0;
    blank_secs = DEFAULT_BLANK_SECS;
#ifndef NO_POWERCTL
    power_secs = DEFAULT_POWER_SECS;
#endif /* NO_POWERCTL */
    num_fbnames = 0;
    fbnames[0] = 0;
    while ( argn < argc && argv[argn][0] == '-' )
	{
	if ( strncmp( argv[argn], "-delay", max(strlen(argv[argn]),2) ) == 0 )
	    {
	    ++argn;
	    if ( argn >= argc ||
		 sscanf( argv[argn], "%u", &blank_secs ) != 1 )
		{
		(void) fprintf( stderr, usage, argv0 );
		exit( 1 );
		}
	    }
#ifndef NO_POWERCTL
	else if ( strncmp( argv[argn], "-powerdelay", max(strlen(argv[argn]),2) ) == 0 )
	    {
	    ++argn;
	    if ( argn >= argc ||
		 sscanf( argv[argn], "%u", &power_secs ) != 1 )
		{
		(void) fprintf( stderr, usage, argv0 );
		exit( 1 );
		}
	    }
#endif /* NO_POWERCTL */
	else if ( strncmp(argv[argn],"-mouse",max(strlen(argv[argn]),2)) == 0 )
	    ign_mouse = 1;
	else if ( strncmp(argv[argn],"-keyboard",max(strlen(argv[argn]),2))==0 )
	    ign_keyboard = 1;
	else if ( strncmp( argv[argn], "-fb", max(strlen(argv[argn]),2) ) == 0 )
	    {
	    ++argn;
	    if ( argn >= argc )
		{
		(void) fprintf( stderr, usage, argv0 );
		exit( 1 );
		}
	    if ( num_fbnames == MAX_FBS )
		{
		(void) fprintf( stderr, "%s: too many -fb options\n", argv0 );
		exit( 1 );
		}
	    fbnames[num_fbnames] = argv[argn];
	    fbnames[++num_fbnames] = 0;
	    }
	else
	    {
	    (void) fprintf( stderr, usage, argv0 );
	    exit( 1 );
	    }
	++argn;
	}

    if ( argn != argc )
	{
	(void) fprintf( stderr, usage, argv0 );
	exit( 1 );
	}

    find_fbs();

    console_name = CONSOLE_NAME;
    keyboard_name = KEYBOARD_NAME;
    mouse_name = MOUSE_NAME;

    some_opens_failed = 0;
    console_fd = my_open( &console_name );
#ifdef NO_OPENKBMS
    keyboard_fd = mouse_fd = -1;
    some_opens_failed = 1;
#else /* NO_OPENKBMS */
    if ( ! ign_keyboard )
	keyboard_fd = my_open( &keyboard_name );
    if ( ! ign_mouse )
	mouse_fd = my_open( &mouse_name );
#endif /* NO_OPENKBMS */
#ifndef NO_POLL
    if ( some_opens_failed )
	(void) fprintf( stderr,
	    "%s: warning - falling back on less efficient blanking method\n",
	    argv0 );
#endif /* NO_POLL */

    daemonize();
    /* No writing to stderr allowed after this point. */

    (void) signal( SIGHUP, sig_handle );
    (void) signal( SIGINT, sig_handle );
    (void) signal( SIGTERM, sig_handle );

    /* Power-up and unblank at start in case the system crashed while
    ** powered down or blanked and somehow didn't get fixed when rebooting.
    ** It happens, rarely.
    */
#ifndef NO_POWERCTL
    poweron();
#endif /* NO_POWERCTL */
    unblank();

    /* Main loop. */
    main_loop();	/* (never returns) */
    }


static void
find_fbs()
    {
    num_fbs = 0;
    if ( fbnames[0] != (char*) 0 )
	check_fbs( fbnames, 1 );
    else
	{
#ifdef NO_DEVFBS
	static char* builtin_fbnames[] = {
	    "/dev/fb", "/dev/fb0", "/dev/fb1", "/dev/fb2", "/dev/fb3",
	    "/dev/bwone0", "/dev/bwone1",
	    "/dev/bwtwo0", "/dev/bwtwo1",
	    "/dev/cgone0", "/dev/cgone1",
	    "/dev/cgtwo0", "/dev/cgtwo1",
	    "/dev/cgthree0", "/dev/cgthree1",
	    "/dev/cgfour0", "/dev/cgfour1",
	    "/dev/cgfive0", "/dev/cgfive1",
	    "/dev/cgsix0", "/dev/cgsix1", "/dev/cgsix2", "/dev/cgsix3",
	    "/dev/cgeight0", "/dev/cgeight1",
	    "/dev/cgnine0", "/dev/cgnine1",
	    "/dev/cgthirteen0", "/dev/cgthirteen1",
	    "/dev/cgfourteen0", "/dev/cgfourteen1",
	    "/dev/gpone0a", "/dev/gpone1a",
	    "/dev/gt0", "/dev/gt1",
	    "/dev/leo0", "/dev/leo1",
	    (char*) 0
	    };

	check_fbs( builtin_fbnames, 0 );
#else /* NO_DEVFBS */
	char* devfbs_fbnames[MAX_FBS+1];
	int n;
	DIR* d;
	struct dirent* de;
	char file[MAXNAMLEN];
#define DEV_FBS "/dev/fbs"

	d = opendir( DEV_FBS );
	if ( d == (DIR*) 0 )
	    {
	    (void) fprintf(
		stderr, "%s: opendir(\"%s\") - %s\n", argv0, DEV_FBS,
		strerror( errno ) );
	    exit( 1 );
	    }
	n = 0;
	while ( ( de = readdir( d ) ) != (struct dirent*) 0 )
	    {
	    if ( *de->d_name == '.' )
		continue;
	    (void) sprintf( file, "%s/%s", DEV_FBS, de->d_name );
	    devfbs_fbnames[n] = strdup( file );
	    ++n;
	    }
	(void) closedir( d );
	devfbs_fbnames[n] = (char*) 0;

	check_fbs( devfbs_fbnames, 0 );
#endif /* NO_DEVFBS */
	}
    }


static void
check_fbs( names, dodie )
    char** names;
    int dodie;
    {
    char** n;
    int fd, i;
    struct stat sb;

    for ( n = names; *n; ++n )
	{
	fd = open( *n, O_RDWR );
	if ( fd < 0 )
	    {
	    if ( dodie )
		die( *n );
	    }
	else
	    {
	    /* Check for duplicate devices. */
	    if ( fstat( fd, &sb ) != 0 )
		die( *n );
	    for ( i = 0; i < num_fbs; ++i )
		{
		if ( fb_ino[i] == sb.st_ino && fb_dev[i] == sb.st_dev )
		    {
		    (void) close( fd );
		    goto next;
		    }
		}
	    fb_fd[num_fbs] = fd;
	    fbnames[num_fbs] = *n;
	    fb_ino[num_fbs] = sb.st_ino;
	    fb_dev[num_fbs] = sb.st_dev;
	    video_state[num_fbs] = 0;
	    can_setvideo[num_fbs] = 1;
#ifndef NO_POWERCTL
	    power_state[num_fbs] = 0;
	    can_setpower[num_fbs] = 1;
#endif /* NO_POWERCTL */
	    ++num_fbs;
	    }
	next: ;
	}
    }


static void
daemonize()
    {
    int nfds, fd, i;
    struct rlimit rl;

    switch ( fork() )
	{
	case 0:
	    break;
	case -1:
	    die( "fork" );
	default:
	    exit( 0 );
	}

#ifdef RLIMIT_NOFILE
    if ( getrlimit( RLIMIT_NOFILE, &rl ) < 0 )
	die( "getrlimit" );
    nfds = rl.rlim_max;
#else /* RLIMIT_NOFILE */
    nfds = getdtablesize();
#endif /* RLIMIT_NOFILE */

    /* Close all fds. */
    for ( fd = 0; fd < nfds; ++fd )
	{
	/* Except don't close fds we're interested in. */
	if ( fd == console_fd ||
	     ( ! ign_keyboard && fd == keyboard_fd ) ||
	     ( ! ign_mouse && fd == mouse_fd ))
	    goto do_not_close;
	for ( i = 0; i < num_fbs; ++i )
	    if ( fb_fd[i] == fd )
		goto do_not_close;
	(void) close( fd );
	do_not_close: ;
	}

#ifndef NO_SETSID
    (void) setsid();
#endif /* NO_SETSID */
    }


static char*
resolve_symlink( linkname )
    char* linkname;
    {
    /* Try to resolve a symlink once, so repeated stat() calls on
    ** diskless workstations don't drive the server into the ground.
    **/
    int bufsize, linklen;
    char* buf;
    char* p = strrchr( linkname, '/' );
    int dirlen = p ? p + 1 - linkname : 0;

    for ( bufsize = 256; ; bufsize *= 2 )
	if ( dirlen < bufsize )
	    {
	    buf = malloc( bufsize );
	    if ( ! buf )
		{
		(void) fprintf( stderr, "%s: out of memory\n", argv0 );
		exit( 1 );
		}
	    linklen = readlink( linkname, buf + dirlen, bufsize - dirlen );
	    if ( linklen < 0 )
		/* Not a symlink. */
		return linkname;
	    if ( linklen < bufsize - dirlen )
		break;
	    free( buf );
	    }

    buf[dirlen + linklen] = 0;

    if ( buf[dirlen] == '/' )
	/* Absolute symlink. */
	return buf + dirlen;

    /* Relative symlink. */
    (void) strncpy( buf, linkname, dirlen );
    return buf;
    }


static void
main_loop()
    {
    int state;
#define ST_AWAKE 0
#define ST_BLANKED 1
#ifndef NO_POWERCTL
#define ST_POWEROFF 2
#endif /* NO_POWERCTL */
    unsigned idle_secs;
    long sleep_msecs;
#ifndef NO_POLL
    long poll_msecs;
#endif /* NO_POLL */

    state = ST_AWAKE;
#ifdef DEBUG
    syslog( LOG_ALERT, "%s: starting in ST_AWAKE", argv0 );
#endif
    for (;;)
	{
	idle_secs = idle_time();
#ifdef DEBUG
	syslog( LOG_ALERT, "%s: idle_secs=%d", argv0, idle_secs );
#endif

	/* Figure out if we should change states, and if not how long we
	** should sleep or poll for.
	*/
	again: ;
	switch ( state )
	    {
	    case ST_AWAKE:
	    if ( idle_secs >= blank_secs )
		{
		blank();
		state = ST_BLANKED;
#ifdef DEBUG
		syslog( LOG_ALERT, "%s: switching to ST_BLANKED", argv0 );
#endif
		goto again;
		}
	    sleep_msecs = ( blank_secs - idle_secs ) * 1000;
#ifndef NO_POLL
	    poll_msecs = ( blank_secs - idle_secs ) * 1000;
#endif /* NO_POLL */
	    break;

	    case ST_BLANKED:
#ifndef NO_POWERCTL
	    if ( idle_secs >= power_secs )
		{
		poweroff();
		state = ST_POWEROFF;
#ifdef DEBUG
		syslog( LOG_ALERT, "%s: switching to ST_POWEROFF", argv0 );
#endif
		goto again;
		}
#endif /* NO_POWERCTL */
	    if ( idle_secs < blank_secs )
		{
		unblank();
		state = ST_AWAKE;
#ifdef DEBUG
		syslog( LOG_ALERT, "%s: switching to ST_AWAKE", argv0 );
#endif
		goto again;
		}
#ifndef NO_POLL
#ifndef NO_POWERCTL
	    poll_msecs = ( power_secs - idle_secs ) * 1000;
#else /* NO_POWERCTL */
	    poll_msecs = 1000000000;	/* long time */
#endif /* NO_POWERCTL */
#endif /* NO_POLL */
	    sleep_msecs = IDLECHECK_MSECS;
	    break;

#ifndef NO_POWERCTL
	    case ST_POWEROFF:
	    if ( idle_secs < power_secs )
		{
		poweron();
		state = ST_BLANKED;
#ifdef DEBUG
		syslog( LOG_ALERT, "%s: switching back to ST_BLANKED", argv0 );
#endif
		goto again;
		}
#ifndef NO_POLL
	    poll_msecs = 1000000000;	/* long time */
#endif /* NO_POLL */
	    sleep_msecs = IDLECHECK_MSECS;
	    break;
#endif /* NO_POWERCTL */

	    default:
	    syslog( LOG_ERR, "%s: unknown state", argv0 );
	    exit( 1 );
	    }

        /* Now do the sleep or poll. */
#ifndef NO_POLL
	if ( ! some_opens_failed )
	    {
	    /* If possible, we wait for a long time or until one of the fd's
	    ** has data available.  This is a big win over waking up evey 1/4
	    ** second.
	    */
	    struct pollfd pollfd[3];
	    unsigned long nfds = 1;

	    pollfd[0].fd = console_fd;
	    pollfd[0].events = POLLIN|POLLPRI;
	    if ( ! ign_keyboard )
		{
		pollfd[nfds].fd = keyboard_fd;
		pollfd[nfds].events = POLLIN|POLLPRI;
		++nfds;
		}
	    if ( ! ign_mouse )
		{
		pollfd[nfds].fd = mouse_fd;
		pollfd[nfds].events = POLLIN|POLLPRI;
		++nfds;
		}
#ifdef DEBUG
	    syslog( LOG_ALERT, "%s: polling, poll_msecs=%d", argv0, poll_msecs );
#endif
	    (void) poll( pollfd, nfds, poll_msecs );

	    /* Most likely there are now characters waiting to be read,
	    ** but it's possible there is no process reading them.  If
	    ** that's the case then poll() will continue to return
	    ** immediately while idle_time() shows large values, which
	    ** would result in a hard-loop calling poll().  To prevent
	    ** such a loop we always do a minimum 1/4 second sleep.
	    */
	    msleep( IDLECHECK_MSECS );
	    }
	else
	    msleep( sleep_msecs );
#else /* NO_POLL */
#ifdef DEBUG
	syslog( LOG_ALERT, "%s: sleeping, sleep_msecs=%d", argv0, sleep_msecs );
#endif
	msleep( sleep_msecs );
#endif /* NO_POLL */
	}
    }


static unsigned
idle_time()
    {
    struct stat sb;
    unsigned min_time;
    time_t touch_time;
    time_t now;

    my_stat( console_fd, console_name, &sb );
    touch_time = max( sb.st_atime, sb.st_mtime );
    now = time( (time_t*) 0 );
    min_time = now - touch_time;

    if ( ! ign_keyboard )
	{
	my_stat( keyboard_fd, keyboard_name, &sb );
	touch_time = max( sb.st_atime, sb.st_mtime );
	min_time = min( min_time, (unsigned) ( now - touch_time ) );
	}

    if ( ! ign_mouse )
	{
	my_stat( mouse_fd, mouse_name, &sb );
	touch_time = max( sb.st_atime, sb.st_mtime );
	min_time = min( min_time, (unsigned) ( now - touch_time ) );
	}

    return min_time;
    }


static void
msleep( msecs )
    long msecs;
    {
    struct timeval timeout;

    timeout.tv_sec = msecs / 1000;
    timeout.tv_usec = ( msecs % 1000 ) * 1000;
    (void) select( 0, NULL, NULL, NULL, &timeout );
    }


static int
my_open( name )
    char** name;
    {
    int fd;

    fd = open( *name, O_RDONLY );
    if ( fd < 0 )
	{
	(void) fprintf(
	    stderr, "%s: %s - %s\n", argv0, *name, strerror( errno ) );
	some_opens_failed = 1;
	*name = resolve_symlink( *name );
	}
    return fd;
    }


static void
my_stat( fd, dev, sbP )
    int fd;
    char* dev;
    struct stat* sbP;
    {
    if ( ( fd >= 0 ? fstat( fd, sbP ) : stat( dev, sbP ) ) != 0 )
	{
	syslog( LOG_ERR, "%s: stat %s - %m", argv0, dev );
#ifndef NO_POWERCTL
	poweron();
#endif /* NO_POWERCTL */
	unblank();
	exit( 1 );
	}
    }


static void
blank()
    {
    int fb_num;

    for ( fb_num = 0; fb_num < num_fbs; ++fb_num )
	if ( can_setvideo[fb_num] && video_state[fb_num] )
	    {
	    set_video( fb_num, 0 );
	    video_state[fb_num] = 0;
	    }
    }


static void
unblank()
    {
    int fb_num;

    for ( fb_num = 0; fb_num < num_fbs; ++fb_num )
	if ( can_setvideo[fb_num] && ! video_state[fb_num] )
	    {
	    set_video( fb_num, 1 );
	    video_state[fb_num] = 1;
	    }
    }


static void
set_video( fb_num, b )
    int fb_num, b;
    {
    int fd, fbv;

    fd = fb_fd[fb_num];
    fbv = b ? FBVIDEO_ON : FBVIDEO_OFF;
    if ( ioctl( fd, FBIOSVIDEO, &fbv ) < 0 )
	{
	(void) syslog( LOG_ERR,
	    "%s: error setting video on %s - %m", argv0, fbnames[fb_num] );
	can_setvideo[fb_num] = 0;
	}
    }


#ifndef NO_POWERCTL

static void
poweroff()
    {
    int fb_num;

    for ( fb_num = 0; fb_num < num_fbs; ++fb_num )
	{
	if ( can_setpower[fb_num] && power_state[fb_num] )
	    {
	    set_power( fb_num, 0 );
	    power_state[fb_num] = 0;
	    }
	}
    }


static void
poweron()
    {
    int fb_num;

    for ( fb_num = 0; fb_num < num_fbs; ++fb_num )
	if ( can_setpower[fb_num] && ! power_state[fb_num] )
	    {
	    set_power( fb_num, 1 );
	    power_state[fb_num] = 1;
	    }
    }


static void
set_power( fb_num, b )
    int fb_num, b;
    {
    int fd, fbv;

    fd = fb_fd[fb_num];
    fbv = b ? 1 : 0;
    if ( ioctl( fd, SET_MONITOR_POWER, &fbv ) < 0 )
	{
	if ( errno != ENOTTY )
	    (void) syslog( LOG_ERR,
		"%s: error setting power on %s - %m", argv0, fbnames[fb_num] );
	can_setpower[fb_num] = 0;
	}
    }

#endif /* NO_POWERCTL */


static void
sig_handle()
    {
#ifndef NO_POWERCTL
    poweron();
#endif /* NO_POWERCTL */
    unblank();
    exit( 0 );
    }


static void
die( name )
    char* name;
    {
    (void) fprintf( stderr, "%s: %s - %s\n", argv0, name, strerror( errno ) );
    exit( 1 );
    }


#ifdef NO_STRERROR
static char*
strerror( e )
    int e;
    {
    extern int sys_nerr;
    extern char* sys_errlist[];
    return e >= sys_nerr ? "unknown error" : sys_errlist[e];
    }
#endif /* NO_STRERROR */
